"""SPDX file processor package"""
from .processor import SPDXProcessor
from .models import (
    Checksum, CreationInfo, File, ExternalDocumentRef,
    Package, Relationship, SystemRoot, Document, IndexRoot
)

__version__ = "0.1.4"
__author__ = "Dinesh Ravi"



from typing import List, Any
from dataclasses import dataclass

@dataclass
class Checksum:
    algorithm: str
    checksumValue: str

    @staticmethod
    def from_dict(obj: Any) -> 'Checksum':
        _algorithm = str(obj.get("algorithm"))
        _checksumValue = str(obj.get("checksumValue"))
        return Checksum(_algorithm, _checksumValue)

@dataclass
class CreationInfo:
    comment: str
    created: str
    creators: List[str]
    licenseListVersion: str

    @staticmethod
    def from_dict(obj: Any) -> 'CreationInfo':
        if not isinstance(obj, dict):
            raise TypeError(f"Expected dict, but got {type(obj).__name__}: {obj}")

        _comment = str(obj.get("comment", ""))
        _created = str(obj.get("created", ""))
        _creators = [str(y) for y in obj.get("creators", [])]
        _licenseListVersion = str(obj.get("licenseListVersion", ""))

        return CreationInfo(_comment, _created, _creators, _licenseListVersion)

@dataclass
class File:
    SPDXID: str
    checksums: List[Checksum]
    copyrightText: str
    fileName: str
    fileTypes: List[str]
    licenseConcluded: str
    licenseInfoInFiles: List[str]

    @staticmethod
    def from_dict(obj: Any) -> 'File':
        _SPDXID = str(obj.get("SPDXID"))
        _checksums = [Checksum.from_dict(y) for y in obj.get("checksums")]
        _copyrightText = str(obj.get("copyrightText"))
        _fileName = str(obj.get("fileName"))
        _fileTypes = [str(y) for y in obj.get("fileTypes",[])]
        _licenseConcluded = str(obj.get("licenseConcluded"))
        _licenseInfoInFiles = [str(y) for y in obj.get("licenseInfoInFiles",[])]
        return File(_SPDXID, _checksums, _copyrightText, _fileName, _fileTypes, _licenseConcluded, _licenseInfoInFiles)

@dataclass
class ExternalDocumentRef:
    checksum: Checksum
    externalDocumentId: str
    spdxDocument: str

    @staticmethod
    def from_dict(obj: Any) -> 'ExternalDocumentRef':
        _checksum = Checksum.from_dict(obj.get("checksum"))
        _externalDocumentId = str(obj.get("externalDocumentId"))
        _spdxDocument = str(obj.get("spdxDocument"))
        return ExternalDocumentRef(_checksum, _externalDocumentId, _spdxDocument)

@dataclass
class Package:
    SPDXID: str
    copyrightText: str
    downloadLocation: str
    licenseConcluded: str
    licenseDeclared: str
    licenseInfoFromFiles: List[str]
    name: str
    supplier: str
    versionInfo: str

    @staticmethod
    def from_dict(obj: Any) -> 'Package':
        if not isinstance(obj, dict):
            raise TypeError(f"Expected dict, but got {type(obj).__name__}: {obj}")

        _SPDXID = str(obj.get("SPDXID", ""))
        _copyrightText = str(obj.get("copyrightText", ""))
        _downloadLocation = str(obj.get("downloadLocation", ""))
        _licenseConcluded = str(obj.get("licenseConcluded", ""))
        _licenseDeclared = str(obj.get("licenseDeclared", ""))
        _licenseInfoFromFiles = [str(y) for y in obj.get("licenseInfoFromFiles", [])]
        _name = str(obj.get("name", ""))
        _supplier = str(obj.get("supplier", ""))
        _versionInfo = str(obj.get("versionInfo", ""))

        return Package(
            _SPDXID, _copyrightText, _downloadLocation, _licenseConcluded,
            _licenseDeclared, _licenseInfoFromFiles, _name, _supplier, _versionInfo
        )

@dataclass
class Relationship:
    relatedSpdxElement: str
    relationshipType: str
    spdxElementId: str
    comment: str

    @staticmethod
    def from_dict(obj: Any) -> 'Relationship':
        _relatedSpdxElement = str(obj.get("relatedSpdxElement"))
        _relationshipType = str(obj.get("relationshipType"))
        _spdxElementId = str(obj.get("spdxElementId"))
        _comment = str(obj.get("comment"))
        return Relationship(_relatedSpdxElement, _relationshipType, _spdxElementId, _comment)

@dataclass
class SystemRoot:
    SPDXID: str
    creationInfo: CreationInfo
    dataLicense: str
    documentNamespace: str
    externalDocumentRefs: List[ExternalDocumentRef]
    name: str
    packages: List[Package]
    relationships: List[Relationship]
    spdxVersion: str

    @staticmethod
    def from_dict(obj: Any) -> 'SystemRoot':
        _SPDXID = str(obj.get("SPDXID"))
        _creationInfo = CreationInfo.from_dict(obj.get("creationInfo"))
        _dataLicense = str(obj.get("dataLicense"))
        _documentNamespace = str(obj.get("documentNamespace"))
        _externalDocumentRefs = [ExternalDocumentRef.from_dict(y) for y in obj.get("externalDocumentRefs")]
        _name = str(obj.get("name"))
        _packages = [Package.from_dict(y) for y in obj.get("packages")]
        _relationships = [Relationship.from_dict(y) for y in obj.get("relationships")]
        _spdxVersion = str(obj.get("spdxVersion"))
        return SystemRoot(_SPDXID, _creationInfo, _dataLicense, _documentNamespace, _externalDocumentRefs, _name, _packages, _relationships, _spdxVersion)

@dataclass
class Document:
    documentNamespace: str
    filename: str
    sha1: str

    @staticmethod
    def from_dict(obj: Any) -> 'Document':
        _documentNamespace = str(obj.get("documentNamespace"))
        _filename = str(obj.get("filename"))
        _sha1 = str(obj.get("sha1"))
        return Document(_documentNamespace, _filename, _sha1)

@dataclass
class IndexRoot:
    documents: List[Document]

    @staticmethod
    def from_dict(obj: Any) -> 'IndexRoot':
        _documents = [Document.from_dict(y) for y in obj.get("documents")]
        return IndexRoot(_documents)


import argparse
import sys
from .processor import SPDXProcessor
import concurrent.futures


def generate_full_details(processor: SPDXProcessor, output_file: str):
    output = processor.process_full()
    with open(output_file, 'w', encoding='utf-8') as f:
        f.write('\n'.join(output))


def generate_binary_details(processor, output_file):
    output = processor.process_binaries()
    with open(output_file, 'w', encoding='utf-8') as f:
        f.write('\n'.join(output))


def generate_minmal_details(processor, output_file):
    output = processor.process_minimal()
    with open(output_file, 'w', encoding='utf-8') as f:
        f.write('\n'.join(output))


def main():
    parser = argparse.ArgumentParser(
        description='Process SPDX files with different output formats')

    parser.add_argument(
        '--archive', default='system_extra.spdx.tar.zst', help='Path to the SPDX archive. defaults to system_extra.spdx.tar.zst')
    # Allow multiple options
    parser.add_argument('-m', '--min', action='store_true',
                        help='Generate minimal output ex:1non_build_recipes_only.txt')
    parser.add_argument('-b', '--bin', action='store_true',
                        help='Generate binary-focused output ex:2non_build_binaries_only.txt')
    parser.add_argument('-f', '--full', action='store_true',
                        help='Generate full output ex:3with_build_deps_full.txt')

    args = parser.parse_args()

    # If no optional argument is provided, enable all options
    if not (args.min or args.bin or args.full):
        args.min = True
        args.bin = True
        args.full = True

    try:
        processor = SPDXProcessor(args.archive)
        processor.load_data()

        # Handle combined options
        output = ""
        function_mapping = {  # (function, (arg1, arg2))
            "generate_minmal_details": (generate_minmal_details, (processor, '1non_build_recipes_only.txt')),
            "generate_binary_details": (generate_binary_details, (processor, '2non_build_binaries_only.txt')),
            "generate_full_details": (generate_full_details, (processor, '3with_build_deps_full.txt')),
        }

        # Get user choices
        selected_tasks = []
        if args.full:
            # generate_full_details(
            #     processor, output_file='3with_build_deps_full.txt')
            selected_tasks.append(function_mapping["generate_full_details"])
        if args.bin:
            # generate_binary_details(
            #     processor, output_file='2non_build_binaries_only.txt')
            selected_tasks.append(function_mapping["generate_binary_details"])
        if args.min:
            # generate_minmal_details(
            #     processor, output_file='1non_build_recipes_only.txt')
            selected_tasks.append(function_mapping["generate_minmal_details"])
        # Check if no options were selected
        # if not output:
        #     print("Error: No valid output options selected.", file=sys.stderr)
        #     sys.exit(1)

        # Execute selected functions in parallel (CPU bound)
        with concurrent.futures.ProcessPoolExecutor() as executor:
            futures = {executor.submit(
                func, *args): func.__name__ for func, args in selected_tasks}

            for future in concurrent.futures.as_completed(futures):
                if future.result():
                    # Only prints if result is none
                    print(f"{futures[future]} result: {future.result()}")
                else:
                    print(f"{futures[future]} done!")

    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    main()



"""Functions from processIndexbinaries.py"""
import os
import json
import sys
from typing import List
from typing import Any
from ..models import (
    IndexRoot, SystemRoot, Document, ExternalDocumentRef,
    File, Package, Relationship
)

# Global variables for compatibility with original functions
g_indexroot = None
g_sysroot = None
g_root_pkg = None

class BinaryProcessor:
    def __init__(self, input_folder: str, indexroot: IndexRoot = None, sysroot: SystemRoot = None):
        self.input_folder = input_folder
        self.indexroot = indexroot
        self.sysroot = sysroot
        self.root_pkg = None
        if sysroot and sysroot.packages:
            self.root_pkg = sysroot.packages[0]
        
        # Initialize globals for compatibility with original functions
        global g_indexroot, g_sysroot, g_root_pkg
        g_indexroot = self.indexroot
        g_sysroot = self.sysroot
        g_root_pkg = self.root_pkg

    def process(self) -> List[str]:
        # Redirect stdout to capture print output
        from io import StringIO
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        
        try:
            process_system_root_spdx_files(self.input_folder)
            output = mystdout.getvalue().splitlines()
            return output
        finally:
            sys.stdout = old_stdout

def process_system_root_spdx_files(input_folder: str) -> None:
    global g_indexroot, g_sysroot, g_root_pkg
    recipe_count = 0
    dirs = os.listdir(input_folder)
    
    # Load index.json
    for file in dirs:
        if file == "index.json": 
            recipe_count = recipe_count + 1
            file_path = os.path.join(input_folder, file)
            with open(file_path, "r", encoding="utf-8") as f:
                myjsonstring = json.load(f)
                g_indexroot = IndexRoot.from_dict(myjsonstring)
    
    # Load system_extra.json
    for file in dirs:
        if file.startswith("system_extra") and file.endswith(".json"):
            recipe_count = recipe_count + 1
            file_path = os.path.join(input_folder, file)
            with open(file_path, "r", encoding="utf-8") as f:
                myjsonstring = json.load(f)
                g_sysroot = SystemRoot.from_dict(myjsonstring)

    g_root_pkg = g_sysroot.packages[0]
    sys_rels = g_sysroot.relationships
    indexroot_documents = g_indexroot.documents
    sys_ext_doc_refs = g_sysroot.externalDocumentRefs
    print(f"Machine: {g_root_pkg.name} v{g_root_pkg.versionInfo}")

    root_rel_types = sorted(set([rel.relationshipType for rel in sys_rels]))

    for root_rel_type in root_rel_types:
        if root_rel_type not in ["DESCRIBES", "AMENDS", "OTHER"]:
            print(f"{root_rel_type}:")
            rel_count = 0
            rel_count2 = 0
            for sys_rel in sys_rels:
                if sys_rel.relationshipType == root_rel_type:
                    if "DocumentRef-runtime" not in sys_rel.relatedSpdxElement:
                        rel_count = rel_count + 1
                        doc_ref = sys_rel.relatedSpdxElement.split(":SPDXRef-Package")[0]
                        file_name = find_and_get_filename(doc_ref, sys_ext_doc_refs, indexroot_documents)
                        print(f"   {rel_count}.  ------")
                        process_pkg_spdx_files(input_folder, file_name, indexroot_documents)
                    
                    if "DocumentRef-runtime" in sys_rel.relatedSpdxElement:
                        rel_count2 = rel_count2 + 1
                        doc_ref_runtime = sys_rel.relatedSpdxElement.split(":SPDXRef-DOCUMENT")[0]
                        file_name = find_and_get_filename(doc_ref_runtime, sys_ext_doc_refs, indexroot_documents)
                        print(f"   {rel_count2}. {file_name} | {sys_rel.comment}")
                        process_runtime_spdx_files(input_folder, file_name, indexroot_documents)

def process_runtime_spdx_files(input_folder: str, input_file: str, indexroot_documents: List[Document]) -> None:
    file_path = os.path.join(input_folder, input_file)
    with open(file_path, "r", encoding="utf-8") as f:
        data = json.load(f)
        relationships = data.get("relationships", [])
        externalDocumentRefs = [ExternalDocumentRef.from_dict(ref) for ref in data.get("externalDocumentRefs", [])]
        relationship_types = set([rel["relationshipType"] for rel in relationships])
        
        for relationship_type in relationship_types:
            if relationship_type not in ["DESCRIBES", "AMENDS"]:
                print(f"    {relationship_type}:")
                rel_count = 0
                for relationship in relationships:
                    if relationship["relationshipType"] == relationship_type:
                        rel_count = rel_count + 1
                        doc_ref_runtime = relationship["spdxElementId"].split(":")[0]
                        file_name = find_and_get_filename(doc_ref_runtime, externalDocumentRefs, indexroot_documents)
                        process_pkg_spdx_files(input_folder, file_name, indexroot_documents)

def getfilename_from_index(spdxDocument_link: str, indexroot_documents: List[Document]) -> str:
    for doc in indexroot_documents:
        if doc.documentNamespace == spdxDocument_link:
            return doc.filename
    return ""

def find_and_get_filename(dec_ref: str, sys_ext_doc_refs: List[ExternalDocumentRef], indexroot_documents: List[Document]) -> str:
    for ref in sys_ext_doc_refs:
        if ref.externalDocumentId == dec_ref:
            return getfilename_from_index(ref.spdxDocument, indexroot_documents)
    return ""

def process_recipe_spdx_files(input_folder: str, input_file: str, indexroot_documents: List[Document]) -> None:
    file_path = os.path.join(input_folder, input_file)
    with open(file_path, "r", encoding="utf-8") as f:
        data = json.load(f)
        pkg = data.get("packages", [{}])[0]
        supplier = pkg.get("supplier", "Unknown")
        description = pkg.get("description", "Unknown")
        summary = pkg.get("summary", "Unknown")
        downloadLocation = pkg.get("downloadLocation", "Unknown")
        homepage = pkg.get("homepage", "Unknown")
        
        if supplier != "Unknown":
            print(f"                 supplier         : {supplier}")
        if description != "Unknown":
            print(f"                 description      : {description}")
        if summary != "Unknown":
            print(f"                 summary          : {summary}")
        if downloadLocation != "Unknown":
            print(f"                 downloadLocation : {downloadLocation}")
        if homepage != "Unknown":
            print(f"                 homepage         : {homepage}")

def process_recipe_minimal_spdx_info(input_folder: str, input_file: str, indexroot_documents: List[Document]) -> str:
    file_path = os.path.join(input_folder, input_file)
    with open(file_path, "r", encoding="utf-8") as f:
        data = json.load(f)
        pkg = data.get("packages", [{}])[0]
        package_name = pkg.get("name", "Unknown")
        versionInfo = pkg.get("versionInfo", "Unknown")
        downloadLocation = pkg.get("downloadLocation", "Unknown")
        return f"{package_name} : {versionInfo} : {downloadLocation}"

def process_pkg_spdx_files(input_folder: str, input_file: str, indexroot_documents: List[Document]) -> None:
    if not any(input_file.startswith(x) for x in ["runtime-", "recipe-", "index", "system"]) and input_file.endswith(".json"):
        file_path = os.path.join(input_folder, input_file)
        with open(file_path, "r", encoding="utf-8") as f:
            data = json.load(f)
            pkg = data.get("packages", [{}])[0]
            package_name = pkg.get("name", "Unknown")
            versionInfo = pkg.get("versionInfo", "Unknown")
            
            relationships = data.get("relationships", [])
            files = [File.from_dict(f) for f in data.get("files", [])]
            externalDocumentRefs = [ExternalDocumentRef.from_dict(ref) for ref in data.get("externalDocumentRefs", [])]
            relationship_types = sorted(set([rel["relationshipType"] for rel in relationships]))
            
            print(f"        {package_name}:{versionInfo}")
            
            for relationship_type in relationship_types:
                if relationship_type not in ["DESCRIBES", "AMENDS"]:
                    print(f"            {relationship_type}:")
                    rel_count = 0
                    for relationship in relationships:
                        if relationship["relationshipType"] == relationship_type:
                            rel_count = rel_count + 1
                            relatedSpdxElement = relationship["relatedSpdxElement"]
                            if relationship_type == "GENERATED_FROM":
                                doc_ref_recipe = relatedSpdxElement.split(":")[0]
                                recipe_file_name = find_and_get_filename(doc_ref_recipe, externalDocumentRefs, indexroot_documents)
                                process_recipe_spdx_files(input_folder, recipe_file_name, indexroot_documents)
                            else:
                                actual_file = find_actual_file(relatedSpdxElement, files)
                                print(f"                 {rel_count}. {actual_file}")
            print('        ------')

def find_actual_file(relatedSpdxElement: str, files: List[File]) -> str:
    for file in files:
        if file.SPDXID == relatedSpdxElement:
            return f"{file.fileName} of type {file.fileTypes[0]}"
    return "Unknown file"


import os
import json
import tarfile
import zstandard as zstd
from typing import List, Optional
from .models import (
    IndexRoot, SystemRoot, Document, ExternalDocumentRef,
    File, Package, Relationship
)
from .processors.full_processor import FullProcessor
from .processors.binary_processor import BinaryProcessor
from .processors.minimal_processor import MinimalProcessor

class SPDXProcessor:
    def __init__(self, archive_path: str = "system_extra.spdx.tar.zst"):
        self.archive_path = archive_path
        self.input_folder = "system_extra.spdx"
        self.indexroot: Optional[IndexRoot] = None
        self.sysroot: Optional[SystemRoot] = None
        self.root_pkg: Optional[Package] = None
        
    def extract_archive(self):
        """Extract the SPDX archive if it hasn't been extracted yet"""
        if not os.path.exists(self.input_folder):
            if not os.path.exists(self.archive_path):
                raise FileNotFoundError(f"Archive file not found: {self.archive_path}")
            
            os.makedirs(self.input_folder, exist_ok=True)
            with open(self.archive_path, 'rb') as fh:
                dctx = zstd.ZstdDecompressor()
                with dctx.stream_reader(fh) as reader:
                    with tarfile.open(fileobj=reader, mode='r|*') as tar:
                        tar.extractall(path=self.input_folder)

    def load_data(self):
        """Load SPDX data from extracted files"""
        if not os.path.exists(self.input_folder):
            self.extract_archive()
            
        dirs = os.listdir(self.input_folder)
        
        # Load index.json
        if "index.json" in dirs:
            with open(os.path.join(self.input_folder, "index.json"), "r", encoding="utf-8") as f:
                self.indexroot = IndexRoot.from_dict(json.load(f))
        
        # Load system_extra.json
        for file in dirs:
            if file.startswith("system_extra") and file.endswith(".json"):
                with open(os.path.join(self.input_folder, file), "r", encoding="utf-8") as f:
                    self.sysroot = SystemRoot.from_dict(json.load(f))
                break
        
        if self.sysroot and self.sysroot.packages:
            self.root_pkg = self.sysroot.packages[0]

    def process_full(self) -> List[str]:
        """Process data using full processor (from processIndex.py)"""
        self.load_data()
        processor = FullProcessor(self.input_folder, self.indexroot, self.sysroot)
        return processor.process()

    def process_binaries(self) -> List[str]:
        """Process data using binary processor (from processIndexbinaries.py)"""
        self.load_data()
        processor = BinaryProcessor(self.input_folder, self.indexroot, self.sysroot)
        return processor.process()

    def process_minimal(self) -> List[str]:
        """Process data using minimal processor (from processminimalIndex.py)"""
        self.load_data()
        processor = MinimalProcessor(self.input_folder, self.indexroot, self.sysroot)
        return processor.process()

    def getfilename_from_index(self, spdxDocument_link: str) -> Optional[str]:
        """Get filename from document namespace in index"""
        if not self.indexroot:
            return None
        for doc in self.indexroot.documents:
            if doc.documentNamespace == spdxDocument_link:
                return doc.filename
        return None

    def find_and_get_filename(self, dec_ref: str) -> Optional[str]:
        """Find filename from external document reference"""
        if not (self.sysroot and self.indexroot):
            return None
        for ref in self.sysroot.externalDocumentRefs:
            if ref.externalDocumentId == dec_ref:
                return self.getfilename_from_index(ref.spdxDocument)
        return None


"""SPDX processors package"""
from .full_processor import FullProcessor
from .binary_processor import BinaryProcessor
from .minimal_processor import MinimalProcessor

__all__ = ['FullProcessor', 'BinaryProcessor', 'MinimalProcessor']